#Brian Giera
#ChE210D
#final project
#import python modules
import numpy as np
import time

#import compiled fortran library
import caplib

import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomwrite



#========== GLOBAL VARIABLES ==========

#set the random number seed; useful for debugging
#np.random.seed(342324) ###Remove this eventually!!!
N=100
L=10
#distance cutoff for pairwise interactions
Cut = L*0.5
dt=0.0001
T=3
 
#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def CalcEnergyForces(Pos, Plates, q_plate, Forces):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global L, Cut
    return caplib.calcenergyforces(Pos, Plates, q_plate, L, Cut, Forces)

        
def InitCapPositions(L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Plates: (200,3) array of capacitor positions
"""
    #compute integer grid # of locations for cubic lattice
    NLat = 10
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    plate_z=np.max(r)+LatSpac
    #make the position array of the capacitor plates' atoms
    Plates=np.zeros((2*len(r)**2,3),float)
    #Empty vector to be filled with dummy names of the capacitor plates
    Names=[]    
    j=0
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    #for x in 
    for x in r:
        for y in r:
            #The Plates positions are determined here such that the plates span the x and y dimensions of the entire box
            Plates[j]=np.array([x,y,plate_z],float) #Top plate will be positively charged (even index)
            j+=1
            Names.append('H')
            Plates[j]=np.array([x,y,-plate_z],float) #Bottom plate will be negatively charged (odd index)
            j+=1
            Names.append('C')
                     
    return Plates,Names

    
def InitIonPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #Empty vector to be filled with dummy names of the capacitor plates
    Names=[]  
    #compute integer grid # of locations for cubic lattice
    L=L/1.5
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                if (i % 2) == 0: #This ion is negative (odd index)
                    Names.append('O')
                else: #This ion is positive (even index)
                    Names.append('N')
                #add a random offset to help initial minimization
                #Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                #Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos,Names
    return Pos,Names


def Berendsen(Vel,T,Tinst,dt):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    Tau=dt/0.01 #calc and justify
    Vel = Vel - Vel.mean(axis=0)
    Lambda = np.sqrt(1+(dt/Tau)*(T/Tinst - 1))
    Vel = Vel * Lambda
    return Vel  


def InitVelocities(N, T, dt):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = Berendsen(Vel, T, InstTemp(Vel),dt)
    return Vel


def InitAccel(Pos,Plates,q_plate):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    PEnergy, Accel = CalcEnergyForces(Pos, Plates, q_plate, np.zeros_like(Pos))
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

def Histogram(Pos,L,m):
    """Returns a histogram of the z-positions for ions
Input:
    Pos: (N,3) array of atomic positions
    L: Box length
    m: number of bins
Output:
    z_pos: Histogram of z positions for the positive ions
    z_neg: Histogram of z positions for the negative ions
"""
    dz=L/float(m) #bin width
    bin=[((-L/2)+i*dz) for i in range(m+1)]
    z_pos = np.zeros((len(bin)-1), int)
    z_neg= np.zeros((len(bin)-1), int)
   
    i=1 #even or odd counter

    for x in Pos: #Scroll through all the ion positions of Pos
        i+=1
        if (i % 2) == 0: #For negative ions (odd index)
            k=0
            while k<(len(bin)-1): #iterate through all the bin sets
                if x[2]>=bin[k] and x[2]<bin[k+1]: #check to see if the z-position is within a particular bin
                    z_neg[k]+=1 # then add a count in bin k
                k+=1 #iterare the bin index
        else:    #Do the same thing for positive ions (positive index)
            k=0
            while k<(len(bin)-1):
                if x[2]>=bin[k] and x[2]<bin[k+1]:
                    z_pos[k]+=1
                k+=1

    return bin,z_pos,z_neg


##########################################################
#Brian Giera
#ChE210D
#Final Project

#Initial Values
q_plate=0
(Plates,CapNames)=InitCapPositions(L)
(Pos,IonNames)=InitIonPositions(N, L)
Vel=InitVelocities(N, T, dt)
Accel=InitAccel(Pos, Plates, q_plate)

#Histogram Information
HistUpdate=100
t_collect=int(5/dt) #Start generating histograms after this time
bins=20 #Number of bins in the z-position histogram
(bin,z_pos,z_neg)=Histogram(Pos,L,bins)
pos_hist=np.zeros_like(z_pos) #Empty vector to fill with histogram values 
neg_hist=np.zeros_like(z_neg) #Empty vector to fill with histogram values


#Video Information
PdbFreq = 300
if UseVisual:
    #Creates a file for ion positions that is updated over time
    PdbI = atomwrite.pdbfile("Ions.pdb", L, AtomNames = IonNames)
    PdbI.write(Pos)

    #Creates and closes a file for wall positions that never change
    PdbC = atomwrite.pdbfile("Plates.pdb", L, AtomNames = CapNames)
    PdbC.write(Plates)
    PdbC.close()


#Adjustable parameters
t_total=int(25/dt)   #Run for this much time (25 dimensionless seconds)
t_charge=int(10/dt) #Turn the charges on at this time (10 dimensionless seconds)
chargevalue=1   #Give the plate particles a charge


#Various iterators
frame=0
t=0
p=0
h=0
HistCount=0

print "This simulation is for %i particles, T = %i, plates charged with %i, total time = %.1f" % (N,T,chargevalue,int(t_total*dt))

#Print bin dimensions for data manipulation convenience
print "All histograms have the following bin dimensions:"
for x in pos_hist:
    print "[%1.1f,%1.1f)" % (bin[h],bin[h+1])
    h+=1
h=0


while t<=t_total:
     #Vel=Berendsen(Vel, T,InstTemp(Vel),dt) #Rescale velos via Berendsen
     #(Pos,Vel,Accel,KEnergy,PEnergy)=caplib.vvintegrate(Pos, Plates, Vel, Accel, q_plate, L, Cut, dt)
     
     if (t==t_charge):
         q_plate=chargevalue #Give the plate particles a charge
         print "\n\nThe charge was turned on at frame = %i.\n\n" % (frame)

     if t==(t_collect+int(1/dt)):
        #Create a density histograms
        pos_hist=pos_hist/(HistCount*L*L*L/float(bins)) 
        neg_hist=neg_hist/(HistCount*L*L*L/float(bins))
         
        #Print the histograms for this time segment
        print "\nPlate charge = %i\nPositive ion histogram for %i <= time < %i." % (q_plate,int(dt*t_collect),int(dt*(t_collect+int(1/dt))))        
        for x in pos_hist:
           print "%2.5f" % (x)

        print "Plate charge = %i\nNegative ion histogram for %i <= time < %i." % (q_plate,int(dt*t_collect),int(dt*(t_collect+int(1/dt))))
        for x in neg_hist:
           print "%2.5f" % (x)

        t_collect+=int(1/dt) #Reset the segment of time for the next histogram
        HistCount=0 #Reset the amount of data points collected for next time segment
        h=0

     elif (t>=t_collect): #This generates a new histogram that span 1 second each
        h+=1      
        if (h % HistUpdate)==0:
            (bin,z_pos,z_neg)=Histogram(Pos,L,bins)
            pos_hist=pos_hist+z_pos
            neg_hist=neg_hist+z_neg
            HistCount+=1
            h=0

     p+=1
     if (p % PdbFreq)==0 and UseVisual:
        PdbI.write(Pos)
        p=0
        frame+=1

     t+=1 #Increase the time step

if UseVisual:
    PdbI.close()
